@@ -12,15 +12,17 @@ use Digest::MD5::File qw(file_md5 file_md5_hex);
 use English qw{-no_match_vars};
 use File::stat;
 use IO::File;
+use IO::Scalar;
 use MIME::Base64;
-use XML::LibXML;
+use Scalar::Util qw{reftype};
 use URI;
 
 use parent qw{Class::Accessor::Fast};
 
-our $VERSION = '0.54'; ## no critic
+our $VERSION = '0.55'; ## no critic
 
-__PACKAGE__->mk_accessors(qw{bucket creation_date account buffer_size});
+__PACKAGE__->mk_accessors(
+  qw{bucket creation_date account buffer_size region logger verify_region });
 
 ########################################################################
 sub new {
@@ -38,6 +40,27 @@ sub new {
   croak 'no account'
     if !$self->account;
 
+  if ( !$self->logger ) {
+    $self->logger( $self->account->get_logger );
+  }
+
+  # now each bucket maintains its own region
+  if ( !$self->region && $self->verify_region ) {
+    my $region;
+
+    if ( !$self->account->err ) {
+      $region = $self->get_location_constraint() // 'us-east-1';
+    }
+
+    $self->logger->debug( sprintf "bucket: %s region: %s\n",
+      $self->bucket, ( $region // $EMPTY ) );
+
+    $self->region($region);
+  }
+  elsif ( !$self->region ) {
+    $self->region( $self->account->region );
+  }
+
   return $self;
 } ## end sub new
 
@@ -46,6 +69,10 @@ sub _uri {
 ########################################################################
   my ( $self, $key ) = @_;
 
+  if ($key) {
+    $key =~ s/^\///xsm;
+  }
+
   my $uri
     = ($key)
     ? $self->bucket . $SLASH . $self->account->_urlencode($key)
@@ -98,20 +125,64 @@ sub add_key {
     $conf->{'Content-MD5'} = $md5_base64;
   } ## end else [ if ( ref $value eq 'SCALAR')]
 
-  # If we're pushing to a bucket that's under DNS flux, we might get a 307
-  # Since LWP doesn't support actually waiting for a 100 Continue response,
-  # we'll just send a HEAD first to see what's going on
+  # If we're pushing to a bucket that's under
+  # DNS flux, we might get a 307 Since LWP doesn't support actually
+  # waiting for a 100 Continue response, we'll just send a HEAD first
+  # to see what's going on
+  my $retval = eval {
+    return $self->_add_key(
+      { headers => $conf,
+        data    => $value,
+        key     => $key
+      }
+    );
+  };
+
+  # one more try? if someone specified the wrong region, we'll get a
+  # 301 and you'll only know the region of redirection - no location
+  # header provided...
+  if ($EVAL_ERROR) {
+    my $rsp = $self->account->last_response;
+    if ( $rsp->code eq '301' ) {
+      $self->region( $rsp->headers->{'x-amz-bucket-region'} );
+    }
+
+    return $self->_add_key(
+      { headers => $conf,
+        data    => $value,
+        key     => $key
+      }
+    );
+  }
 
-  if ( ref $value ) {
-    return $self->account->_send_request_expect_nothing_probed( 'PUT',
-      $self->_uri($key), $conf, $value );
-  } ## end if ( ref $value )
-  else {
-    return $self->account->_send_request_expect_nothing( 'PUT',
-      $self->_uri($key), $conf, $value );
-  } ## end else [ if ( ref $value ) ]
 } ## end sub add_key
 
+sub _add_key {
+  my ( $self, @args ) = @_;
+
+  my ( $data, $headers, $key ) = @{ $args[0] }{qw{data headers key}};
+
+  if ( ref $data ) {
+    return $self->account->_send_request_expect_nothing_probed(
+      { method  => 'PUT',
+        path    => $self->_uri($key),
+        headers => $headers,
+        data    => $data,
+        region  => $self->region,
+      }
+    );
+  } ## end if ( ref $value )
+  else {
+    return $self->account->_send_request_expect_nothing(
+      { method  => 'PUT',
+        path    => $self->_uri($key),
+        headers => $headers,
+        data    => $data,
+        region  => $self->region,
+      }
+    );
+  }
+} ## end else [ if ( ref $value ) ]
 ########################################################################
 sub add_key_filename {
 ########################################################################
@@ -120,12 +191,125 @@ sub add_key_filename {
   return $self->add_key( $key, \$value, $conf );
 } ## end sub add_key_filename
 
+########################################################################
+sub upload_multipart_object {
+########################################################################
+  my ( $self, @args ) = @_;
+
+  my $logger = $self->logger;
+
+  my %parameters;
+
+  if ( @args == 1 && reftype( $args[0] ) eq 'HASH' ) {
+    %parameters = %{ $args[0] };
+  }
+  else {
+    %parameters = @args;
+  }
+
+  croak 'no key!'
+    if !$parameters{key};
+
+  croak 'either data, callback or fh must be set!'
+    if !$parameters{data} && !$parameters{callback} && !$parameters{fh};
+
+  croak 'callback must be a reference to a subroutine!'
+    if $parameters{callback} && reftype( $parameters{callback} ) ne 'CODE';
+
+  $parameters{abort_on_error} //= $TRUE;
+  $parameters{chunk_size}     //= $MIN_MULTIPART_UPLOAD_CHUNK_SIZE;
+
+  if ( !$parameters{callback} && !$parameters{fh} ) {
+    #...but really nobody should be passing a >5MB scalar
+    my $data = ref $parameters{data} ? $parameters{data} : \$parameters{data};
+
+    $parameters{fh} = IO::Scalar->new($data);
+  }
+
+  # ...having a file handle implies, we use this callback
+  if ( $parameters{fh} ) {
+    my $fh = $parameters{fh};
+
+    $fh->seek( 0, 2 );
+
+    my $length = $fh->tell;
+    $fh->seek( 0, 0 );
+
+    $logger->trace( sub { return sprintf 'length of object: %s', $length; } );
+
+    croak 'length of the object must be >= '
+      . $MIN_MULTIPART_UPLOAD_CHUNK_SIZE
+      if $length < $MIN_MULTIPART_UPLOAD_CHUNK_SIZE;
+
+    my $chunk_size
+      = ( $parameters{chunk_size} && $parameters{chunk_size} )
+      > $MIN_MULTIPART_UPLOAD_CHUNK_SIZE
+      ? $parameters{chunk_size}
+      : $MIN_MULTIPART_UPLOAD_CHUNK_SIZE;
+
+    $parameters{callback} = sub {
+      return
+        if !$length;
+
+      my $bytes_read = 0;
+
+      my $n = $length >= $chunk_size ? $chunk_size : $length;
+
+      $logger->trace( sprintf 'reading %d bytes', $n );
+
+      my $buffer;
+
+      my $bytes = $fh->read( $buffer, $n, $bytes_read );
+      $logger->trace( sprintf 'read %d bytes', $bytes );
+
+      $bytes_read += $bytes;
+
+      $length -= $bytes;
+
+      $logger->trace( sprintf '%s bytes left to read', $length );
+
+      return ( \$buffer, $bytes );
+    };
+
+  }
+
+  my $headers = $parameters{headers} || {};
+
+  my $id = $self->initiate_multipart_upload( $parameters{key}, $headers );
+
+  $logger->trace( sprintf 'multipart id: %s', $id );
+
+  my $part = 1;
+  my %parts;
+  my $key = $parameters{key};
+
+  eval {
+    while (1) {
+      my ( $buffer, $length ) = $parameters{callback}->();
+      last if !$buffer;
+
+      my $etag = $self->upload_part_of_multipart_upload(
+        { id => $id, key => $key, data => $buffer, part => $part } );
+
+      $parts{ $part++ } = $etag;
+    }
+
+    $self->complete_multipart_upload( $parameters{key}, $id, \%parts );
+  };
+
+  if ( $EVAL_ERROR && $parameters{abort_on_error} ) {
+    $self->abort_multipart_upload( $key, $id );
+    %parts = ();
+  }
+
+  return \%parts;
+}
+
 # Initiates a multipart upload operation. This is necessary for uploading
 # files > 5Gb to Amazon S3
 #
 # returns: upload ID assigned by Amazon (used to identify this
 # particular upload in other operations)
-
 ########################################################################
 sub initiate_multipart_upload {
 ########################################################################
@@ -136,8 +320,14 @@ sub initiate_multipart_upload {
 
   my $acct = $self->account;
 
-  my $request
-    = $acct->_make_request( 'POST', $self->_uri($key) . '?uploads=', $conf );
+  my $request = $acct->_make_request(
+    { region  => $self->region,
+      method  => 'POST',
+      path    => $self->_uri($key) . '?uploads=',
+      headers => $conf
+    }
+  );
+
   my $response = $acct->_do_http($request);
 
   $acct->_croak_if_response_error($response);
@@ -160,7 +350,27 @@ sub upload_part_of_multipart_upload {
 ########################################################################
   my ( $self, @args ) = @_;
 
-  my ( $key, $upload_id, $part_number, $data, $length ) = @args;
+  my ( $key, $upload_id, $part_number, $data, $length );
+
+  if ( @args == 1 ) {
+    if ( reftype( $args[0] ) eq 'HASH' ) {
+      ( $key, $upload_id, $part_number, $data, $length )
+        = @{ $args[0] }{qw{ key id part data length}};
+    }
+    elsif ( reftype( $args[0] ) eq 'ARRAY' ) {
+      ( $key, $upload_id, $part_number, $data, $length ) = @{ $args[0] };
+    }
+  }
+  else {
+    ( $key, $upload_id, $part_number, $data, $length ) = @args;
+  }
+
+  # argh...wish we didn't have to do this!
+  if ( ref $data ) {
+    $data = ${$data};
+  }
+
+  $length = $length || length $data;
 
   croak 'Object key is required'
     if !$key;
@@ -183,9 +393,18 @@ sub upload_part_of_multipart_upload {
   $conf->{'Content-Length'} = $length;
 
   my $params = "?partNumber=${part_number}&uploadId=${upload_id}";
-  my $request
-    = $acct->_make_request( 'PUT', $self->_uri($key) . $params, $conf,
-    $data );
+
+  $self->logger->debug( 'uploading ' . sprintf 'part: %s length: %s',
+    $part_number, length $data );
+
+  my $request = $acct->_make_request(
+    { region  => $self->region,
+      method  => 'PUT',
+      path    => $self->_uri($key) . $params,
+      headers => $conf,
+      data    => $data
+    }
+  );
 
   my $response = $acct->_do_http($request);
 
@@ -202,6 +421,26 @@ sub upload_part_of_multipart_upload {
   return $etag;
 } ## end sub upload_part_of_multipart_upload
 
+########################################################################
+sub make_xml_document_simple {
+########################################################################
+  my ($parts_hr) = @_;
+
+  my $xml = q{<?xml version="1.0" encoding="UTF-8"?>};
+  my $xml_template
+    = '<Part><PartNumber>%s</PartNumber><ETag>%s</ETag></Part>';
+  my @parts;
+
+  foreach my $part_num ( sort { $a <=> $b } keys %{$parts_hr} ) {
+    push @parts, sprintf $xml_template, $part_num, $parts_hr->{$part_num};
+  }
+
+  $xml .= sprintf "\n<CompleteMultipartUpload>%s</CompleteMultipartUpload>\n",
+    join q{}, @parts;
+
+  return $xml;
+}
+
 #
 # Inform Amazon that the multipart upload has been completed
 # You must supply a hash of part Numbers => eTags
@@ -212,6 +451,8 @@ sub complete_multipart_upload {
 ########################################################################
   my ( $self, $key, $upload_id, $parts_hr ) = @_;
 
+  $self->logger->debug( Dumper( [ $key, $upload_id, $parts_hr ] ) );
+
   croak 'Object key is required'
     if !$key;
 
@@ -224,22 +465,12 @@ sub complete_multipart_upload {
   # The complete command requires sending a block of xml containing all
   # the part numbers and their associated etags (returned from the upload)
 
-  #build XML doc
-  my $xml_doc      = XML::LibXML::Document->new( '1.0', 'UTF-8' );
-  my $root_element = $xml_doc->createElement('CompleteMultipartUpload');
-  $xml_doc->addChild($root_element);
+  # build XML doc
 
-  # Add the content
-  foreach my $part_num ( sort { $a <=> $b } keys %{$parts_hr} ) {
+  my $content = make_xml_document_simple($parts_hr);
 
-    # For each part, create a <Part> element with the part number & etag
-    my $part = $xml_doc->createElement('Part');
-    $part->appendTextChild( 'PartNumber' => $part_num );
-    $part->appendTextChild( 'ETag'       => $parts_hr->{$part_num} );
-    $root_element->addChild($part);
-  } ## end foreach my $part_num ( sort...)
+  $self->logger->debug("content: \n$content");
 
-  my $content    = $xml_doc->toString;
   my $md5        = md5($content);
   my $md5_base64 = encode_base64($md5);
   chomp $md5_base64;
@@ -250,13 +481,24 @@ sub complete_multipart_upload {
     'Content-Type'   => 'application/xml'
   };
 
-  my $acct    = $self->account;
-  my $params  = "?uploadId=${upload_id}";
-  my $request = $acct->_make_request( 'POST', $self->_uri($key) . $params,
-    $conf, $content );
+  my $acct   = $self->account;
+  my $params = "?uploadId=${upload_id}";
+
+  my $request = $acct->_make_request(
+    { region  => $self->region,
+      method  => 'POST',
+      path    => $self->_uri($key) . $params,
+      headers => $conf,
+      data    => $content
+    }
+  );
+
   my $response = $acct->_do_http($request);
 
-  $acct->_croak_if_response_error($response);
+  if ( $response->code !~ /\A2\d\d\z/xsm ) {
+    $acct->_remember_errors( $response->content, 1 );
+    croak $response->status_line;
+  }
 
   return $TRUE;
 } ## end sub complete_multipart_upload
@@ -275,9 +517,16 @@ sub abort_multipart_upload {
   croak 'Upload id is required'
     if !$upload_id;
 
-  my $acct    = $self->account;
-  my $params  = "?uploadId=${upload_id}";
-  my $request = $acct->_make_request( 'DELETE', $self->_uri($key) . $params );
+  my $acct   = $self->account;
+  my $params = "?uploadId=${upload_id}";
+
+  my $request = $acct->_make_request(
+    { region => $self->region,
+      method => 'DELETE',
+      path   => $self->_uri($key) . $params
+    }
+  );
+
   my $response = $acct->_do_http($request);
 
   $acct->_croak_if_response_error($response);
@@ -302,8 +551,14 @@ sub list_multipart_upload_parts {
 
   my $acct   = $self->account;
   my $params = "?uploadId=${upload_id}";
-  my $request
-    = $acct->_make_request( 'GET', $self->_uri($key) . $params, $conf );
+
+  my $request = $acct->_make_request(
+    { region  => $self->region,
+      method  => 'GET',
+      path    => $self->_uri($key) . $params,
+      headers => $conf
+    }
+  );
 
   my $response = $acct->_do_http($request);
 
@@ -323,8 +578,14 @@ sub list_multipart_uploads {
   my ( $self, $conf ) = @_;
 
   my $acct = $self->account;
-  my $request
-    = $acct->_make_request( 'GET', $self->_uri() . '?uploads', $conf );
+
+  my $request = $acct->_make_request(
+    { region  => $self->region,
+      method  => 'GET',
+      path    => $self->_uri() . '?uploads',
+      headers => $conf
+    }
+  );
 
   my $response = $acct->_do_http($request);
 
@@ -357,16 +618,16 @@ sub get_key {
 
   my $uri = $self->_uri($key);
 
-  my $request = $acct->_make_request( $method, $uri, {} );
-
-  my $response = $acct->_do_http( $request, $filename );
-
-  $acct->get_logger->debug(
-    sub {
-      return Dumper( [ $request, $response ] );
+  my $request = $acct->_make_request(
+    { region  => $self->region,
+      method  => $method,
+      path    => $uri,
+      headers => {}
     }
   );
 
+  my $response = $acct->_do_http( $request, $filename );
+
   return if $response->code == 404;
 
   $acct->_croak_if_response_error($response);
@@ -392,7 +653,8 @@ sub get_key {
       ? file_md5_hex($filename)
       : md5_hex( $return->{value} );
 
-    # Some S3-compatible providers return an all-caps MD5 value in the etag so it should be lc'd for comparison.
+    # Some S3-compatible providers return an all-caps MD5 value in the
+    # etag so it should be lc'd for comparison.
     croak "Computed and Response MD5's do not match:  $md5 : $etag"
       if $md5 ne lc $etag;
   } ## end if ( $method eq 'GET' )
@@ -426,8 +688,13 @@ sub delete_key {
   croak 'must specify key'
     if !$key && length $key;
 
-  return $self->account->_send_request_expect_nothing( 'DELETE',
-    $self->_uri($key), {} );
+  return $self->account->_send_request_expect_nothing(
+    { method  => 'DELETE',
+      region  => $self->region,
+      path    => $self->_uri($key),
+      headers => {}
+    }
+  );
 } ## end sub delete_key
 
 ########################################################################
@@ -501,8 +768,13 @@ sub get_acl {
 
   my $acct = $self->account;
 
-  my $request
-    = $acct->_make_request( 'GET', $self->_uri($key) . '?acl=', {} );
+  my $request = $acct->_make_request(
+    { region  => $self->region,
+      method  => 'GET',
+      path    => $self->_uri($key) . '?acl=',
+      headers => {}
+    }
+  );
 
   my $old_redirectable = $acct->ua->requests_redirectable;
   $acct->ua->requests_redirectable( [] );
@@ -516,7 +788,13 @@ sub get_acl {
     my $old_host = $acct->host;
     $acct->host( $uri->host );
 
-    my $request = $acct->_make_request( 'GET', $uri->path, {} );
+    my $request = $acct->_make_request(
+      { region  => $self->region,
+        method  => 'GET',
+        path    => $uri->path,
+        headers => {}
+      }
+    );
 
     $response = $acct->_do_http($request);
 
@@ -553,8 +831,14 @@ sub set_acl {
 
   my $xml = $conf->{acl_xml} || $EMPTY;
 
-  return $self->account->_send_request_expect_nothing( 'PUT', $path,
-    $hash_ref, $xml );
+  return $self->account->_send_request_expect_nothing(
+    { method  => 'PUT',
+      path    => $path,
+      headers => $hash_ref,
+      data    => $xml,
+      region  => $self->region
+    }
+  );
 } ## end sub set_acl
 
 ########################################################################
@@ -562,8 +846,12 @@ sub get_location_constraint {
 ########################################################################
   my ($self) = @_;
 
-  my $xpc
-    = $self->account->_send_request( 'GET', $self->bucket . '/?location=' );
+  my $xpc = $self->account->_send_request(
+    { region => $self->region,
+      method => 'GET',
+      path   => $self->bucket . '/?location='
+    }
+  );
 
   if ( !$xpc ) {
     $self->account->_remember_errors($xpc);
@@ -582,6 +870,14 @@ sub get_location_constraint {
 
 # proxy up the err requests
 
+########################################################################
+sub last_response {
+########################################################################
+  my ($self) = @_;
+
+  return $self->account->last_reponse;
+}
+
 ########################################################################
 sub err {
 ########################################################################
@@ -598,6 +894,14 @@ sub errstr {
   return $self->account->errstr;
 } ## end sub errstr
 
+########################################################################
+sub error {
+########################################################################
+  my ($self) = @_;
+
+  return $self->account->error;
+} ## end sub err
+
 ########################################################################
 sub _content_sub {
 ########################################################################
@@ -652,6 +956,8 @@ sub _content_sub {
 
 __END__
 
+=pod
+
 =head1 NAME
 
 Amazon::S3::Bucket - A container class for a S3 bucket and its contents.
@@ -695,23 +1001,53 @@ Amazon::S3::Bucket - A container class for a S3 bucket and its contents.
 
 Instaniates a new bucket object. 
 
-Requires a hash containing two arguments:
+Pass a hash or hash reference containing various options:
 
 =over
 
-=item bucket
+=item bucket (required)
 
 The name (identifier) of the bucket.
 
-=item account
+=item account (required)
 
 The L<S3::Amazon> object (representing the S3 account) this
 bucket is associated with.
 
+=item buffer_size
+
+The buffer size used for reading and writing objects to S3.
+
+default: 4K
+
+=item region
+
+If no region is set and C<verify_region> is set to true, the region of
+the bucket will be determined by calling the
+C<get_location_constraint> method.  Note that this will decrease
+performance of the constructor. If you know the region or are
+operating in only 1 region, set the region in the C<account> object
+(C<Amazon::S3>).
+
+=item logger
+
+Sets the logger object (should be an object capable of providing at
+least a C<debug> and C<trace> method for recording log messages. If no
+logger object is passed the C<account> object's logger object will be used.
+
+=item verify_region
+
+Indicates that the bucket's region should be determined by calling the
+C<get_location_constraint> method.
+
+default: false
+
 =back
 
-NOTE: This method does not check if a bucket actually
-exists. It simply instaniates the bucket.
+I<NOTE:> This method does not check if a bucket actually exists unless
+you set C<verify_region> to true. If the bucket does not exist,
+the constructor will set the region to the default region specified by
+the L<Amazon::S3> object (C<account>) that you passed.
 
 Typically a developer will not call this method directly,
 but work through the interface in L<S3::Amazon> that will
@@ -719,27 +1055,31 @@ handle their creation.
 
 =head2 add_key
 
-Takes three positional parameters:
+ add_key( key, value, configuration)
+
+Write a new or existing object to S3.
 
 =over
 
 =item key
 
-A string identifier for the resource in this bucket
+A string identifier for the object being written to the bucket.
 
 =item value
 
-A SCALAR string representing the contents of the resource.
+A SCALAR string representing the contents of the object..
 
 =item configuration
 
 A HASHREF of configuration data for this key. The configuration
-is generally the HTTP headers you want to pass the S3
+is generally the HTTP headers you want to pass to the S3
 service. The client library will add all necessary headers.
 Adding them to the configuration hash will override what the
 library would send and add headers that are not typically
 required for S3 interactions.
 
+=item acl_short (optional)
+
 In addition to additional and overriden HTTP headers, this
 HASHREF can have a C<acl_short> key to set the permissions
 (access) of the resource without a seperate call via
@@ -748,8 +1088,16 @@ documentation in C<add_acl> for the values and usage.
 
 =back
 
-Returns a boolean indicating its success. Check C<err> and
-C<errstr> for error message if this operation fails.
+Returns a boolean indicating the sucess or failure of the call. Check
+C<err> and C<errstr> for error messages if this operation fails. To
+examine the raw output of the response from the API call, use the
+C<last_response()> method.
+
+  my $retval = $bucket->add_key('foo', $content, {});
+
+  if ( !$retval ) {
+    print STDERR Dumper([$bucket->err, $bucket->errstr, $bucket->last_response]);
+  }
 
 =head2 add_key_filename
 
@@ -762,6 +1110,20 @@ be streamed rather then loaded into memory in one big chunk.
 Returns a configuration HASH of the given key. If a key does
 not exist in the bucket C<undef> will be returned.
 
+HASH will contain the following members:
+
+=over
+
+=item content_length
+
+=item content_type
+
+=item etag
+
+=item value
+
+=back
+
 =head2 get_key $key_name, [$method]
 
 Takes a key and an optional HTTP method and fetches it from
@@ -824,37 +1186,11 @@ method.
 
 =head2 list_all_v2
 
-List all keys in this bucket without having to worry about
-'marker'. This may make multiple requests to S3 under the
-hood.
+Same as C<list_all> but uses the version 2 API for listing keys.
 
 See L<Amazon::S3/list_bucket_all_v2> for documentation of this
 method.
 
-=head2 abort_multipart_upload
-
-Abort a multipart upload
-
-=head2 complete_multipart_upload
-
-Signal completion of a multipart upload
-
-=head2 initiate_multipart_upload
-
-Initiate a multipart upload
-
-=head2 list_multipart_upload_parts
-
-List all the uploaded parts of a multipart upload
-
-=head2 list_multipart_uploads
-
-List multipart uploads in progress
-
-=head2 upload_part_of_multipart_upload
-
-Upload a portion of a multipart upload
-
 =head2 get_acl
 
 Retrieves the Access Control List (ACL) for the bucket or
@@ -870,7 +1206,9 @@ bucket itself.
 
 =back
 
-=head2 set_acl $conf
+=head2 set_acl
+
+ set_acl(acl)
 
 Retrieves the Access Control List (ACL) for the bucket or
 resource. Requires a HASHREF argument with one of the following keys:
@@ -928,10 +1266,40 @@ Returns a boolean indicating the operations success.
 
 =head2 get_location_constraint
 
-Returns the location constraint data on a bucket.
+Returns the location constraint (region the bucket resides in) for a
+bucket.
+
+Valid values that may be returned:
+
+ af-south-1
+ ap-east-1
+ ap-northeast-1
+ ap-northeast-2
+ ap-northeast-3
+ ap-south-1
+ ap-southeast-1
+ ap-southeast-2
+ ca-central-1
+ cn-north-1
+ cn-northwest-1
+ EU
+ eu-central-1
+ eu-north-1
+ eu-south-1
+ eu-west-1
+ eu-west-2
+ eu-west-3
+ me-south-1
+ sa-east-1
+ us-east-2
+ us-gov-east-1
+ us-gov-west-1
+ us-west-1
+ us-west-2
 
 For more information on location constraints, refer to the
-Amazon S3 Developer Guide.
+documentation for
+L<GetBucketLocation|https://docs.aws.amazon.com/AmazonS3/latest/API/API_GetBucketLocation.html>.
 
 =head2 err
 
@@ -941,6 +1309,165 @@ The S3 error code for the last error the account encountered.
 
 A human readable error string for the last error the account encountered.
 
+=head2 error
+
+The decoded XML string as a hash object of the last error.
+
+=head2 last_response
+
+Returns the last C<HTTP::Response> to an API call.
+
+=head1 MULTIPART UPLOAD SUPPORT
+
+From Amazon's website:
+
+I<Multipart upload allows you to upload a single object as a set of
+parts. Each part is a contiguous portion of the object's data. You can
+upload these object parts independently and in any order. If
+transmission of any part fails, you can retransmit that part without
+affecting other parts. After all parts of your object are uploaded,
+Amazon S3 assembles these parts and creates the object. In general,
+when your object size reaches 100 MB, you should consider using
+multipart uploads instead of uploading the object in a single
+operation.>
+
+See L<https://docs.aws.amazon.com/AmazonS3/latest/userguide/mpuoverview.html> for more information about multipart uploads.
+
+=over 5
+
+=item * Maximum object size 5TB
+
+=item * Maximum number of parts 10,000
+
+=item * Part numbers 1 to 10,000 (inclusive)
+
+=item * Part size 5MB to 5GB. There is no limit on the last part of your multipart upload.
+
+=item * Maximum nubmer of parts returned for a list parts request - 1000
+
+=item * Maximum number of multipart uploads returned in a list multipart uploads request - 1000
+
+=back
+
+A multipart upload begins by calling
+C<initiate_multipart_upload()>. This will return an identifier that is
+used in subsequent calls.
+
+ my $bucket = $s3->bucket('my-bucket');
+ my $id = $bucket->initiate_multipart_upload('some-big-object');
+
+ my $part_list = {};
+
+ my $part = 1;
+ my $etag = $bucket->upload_part_of_multipart_upload('my-bucket', $id, $part, $data, length $data);
+ $part_list{$part++} = $etag;
+
+ $bucket->complete_multipart_upload('my-bucket', $id, $part_list);
+
+=heads upload_multipart_object
+
+ upload_multipart_object( ... )
+
+Convenience routine C<upload_multipart_object> that encapsulates the
+multipart upload process. Accepts a hash or hash reference of
+arguments. If successful, a reference to a hash that contains the part
+numbers and etags of the uploaded parts.
+
+You can pass a data object, callback routine or a file handle.
+
+=over 5
+
+=item key
+
+Name of the key to create.
+
+=item data
+
+Scalar object that contains the data to write to S3.
+
+=item callback
+
+Optionally provided a callback routine that will be called until you
+pass a buffer with a length of 0. Your callback will receive no
+arguments but should return a tuple consisting of a B<reference> to a
+scalar object that contains the data to write and a scalar that
+represents the length of data. Once you return a zero length buffer
+the multipart process will be completed.
+
+=item fh
+
+File handle of an open file. The file must be greater than the minimum
+chunk size for multipart uploads otherwise the method will throw an
+exception.
+
+=item abort_on_error
+
+Indicates whether the multipart upload should be aborted if an error
+is encountered. Amazon will charge you for the storage of parts that
+have been uploaded unless you abort the upload.
+
+default: true
+
+=back
+
+=head2 abort_multipart_upload
+
+ abort_multipart_upload(key, multpart-upload-id)
+
+Abort a multipart upload
+
+=head2 complete_multipart_upload
+
+ complete_multipart_upload(key, multpart-upload-id, parts)
+
+Signal completion of a multipart upload. C<parts> is a reference to a
+hash of part numbers and etags.
+
+=head2 initiate_multipart_upload
+
+ initiate_multipart_upload(key, headers)
+
+Initiate a multipart upload. Returns an id used in subsequent call to
+C<upload_part_of_multipart_upload()>.
+
+=head2 list_multipart_upload_parts
+
+List all the uploaded parts of a multipart upload
+
+=head2 list_multipart_uploads
+
+List multipart uploads in progress
+
+=head2 upload_part_of_multipart_upload
+
+  upload_part_of_multipart_upload(key, id, part, data, length)
+
+Upload a portion of a multipart upload
+
+=over 5
+
+=item key
+
+Name of the key in the bucket to create.
+
+=item id
+
+The multipart-upload id return in the C<initiate_multipart_upload> call.
+
+=item part
+
+The next part number (part numbers start at 1).
+
+=item data
+
+Scalar or reference to a scalar that contains the data to upload.
+
+=item length (optional)
+
+Length of the data.
+
+=back
+
 =head1 SEE ALSO
 
 L<Amazon::S3>
